(in-package :ecocyc)

(defun read-ptools-init-file (&optional (file-path (ptools-init-file-path)) debug?)
  (let* ((file-name (namestring file-path)))
    (init-pt-parameters)
    (if (probe-file file-name)
	(let* ((raw-key-value-list (variable-column-table-to-list
				    :filename file-name
				    :datatypes '(:string :string)
				    ;; By defining the column seps this way we ensure that
				    ;; at most two columns can be parsed in each row, thus
				    ;; allowing param values that contain spaces like filenames.
				    :first-column-sep '(#\Tab #\Space)
				    :column-sep nil
				    :trim-token-chars #\Space
				    ;;kr:Jan-23-2006 This new keyword allows skipping over consecutive blocks of column-separators
				    :no-empty-columns-p t
				    ))
	       (validated-key-value-list)
	       )
	  (format t "[Reading existing Pathway Tools init file: ~S ]~%" file-name)
	  (loop for (key value) in raw-key-value-list
	      for pt-init-param = (find key (get-pt-init-parameters)
					:key #'(lambda (param)
						 (when (pt-init-param-p param)
						   (pt-init-param-key-name param)
						   )
						 )
					:test #'equalp)
	      do
		(when debug?
		  (format *trace-output* "  Param  ~S  <--  ~S~%" key value) )
		(flet ((parse-int-w-warn (string)
			 ;; First perform a simple syntax check.
			 (when (find-if-not #'digit-char-p string)
			   (warn "Parameter ~S in file ~S is not an integer. Instead, it is: ~S"
				 key file-name string)
			   )
			 (parse-dirty-integer-string string)
			 )
		       (parse-eddress-w-warn (string)
			 ;; First perform a simple syntax check.
			 (when (not (find #\@ string))
			   (warn "Parameter ~S in file ~S is not a syntactically correct e-mail address. Instead, it is: ~S"
				 key file-name string)
			   )
			 string
			 )
		       )
		  (when (validate-ptools-init-entry key value)
		    (if pt-init-param
			(let* ((val-type (pt-init-param-value-type pt-init-param))
			       (type-conv-fn
				(case val-type
				  (:string #'identity)
				  (:integer #'parse-int-w-warn)
				  (:eddress #'parse-eddress-w-warn)
				  (:orgid-pub-spec #'parse-www-publish-options)
				  (:list-string    #'parse-www-list-options)
				  (:boolean #'robust-truep)
				  (t (warn "Software error: The value-type for parameter ~S in file ~S has not been defined."
					   value file-name))
				  ) )
			       (converted-value (when type-conv-fn
						  (funcall type-conv-fn value)))
			       (global-var (pt-init-param-ptools-global-var pt-init-param))
			       )
			  ;; This performs the actual initialization by setting the appropriate variable value.
			  (setf (symbol-value global-var) converted-value)
			  (push (list key converted-value) validated-key-value-list)
			  )
		      (warn "Parameter ~S with value ~S in file ~S is not recognized by Pathway Tools. It will be ignored."
			    key value file-name)
		      )
		    )
		  )
		)
	  ;; result:
	  validated-key-value-list
	  )
      (warn "No Pathway Tools init file was found at: ~S" file-name)
      )
    )
  )


(defun variable-column-table-to-list (&key
				      filename
				      (column-sep #\Space)
				      first-column-sep
				      trim-token-chars
				      (comment-chars (list #\; #\#))
				      datatypes
				      no-empty-columns-p
				      with-line-numbers-p
				      null-value
				      )
    (let (row result
	  (row-count 0)
	  ;;kr:Jan-23-2006 This helps to allow more than one column-sep character
	  (all-column-separators (append (listify column-sep) (listify first-column-sep)))
	  )

      (feedinit)

      (with-open-file-noerror (file filename :direction :input)
	(setq result
	  ;; latendre:May-18-2006. replace read-line to read-line-any-platform
	  (loop while (setq row (read-line-any-platform file nil))
	      ;; Do we have a row that contains data?
	      when (and (> (length row) 0)
			(not (loop for i from 0 below (length row)
				 always (find (char row i) all-column-separators)))
			(not (member (char row 0) comment-chars)))
	      collect
		(let* ((row-data (variable-column-row-to-list row column-sep 
							      :datatypes datatypes
							      :no-empty-columns-p no-empty-columns-p
							      :null-value null-value
							      :first-column-sep first-column-sep
							      :trim-token-chars trim-token-chars)))
		  (if with-line-numbers-p
		      (cons (1+ row-count) row-data)
		    row-data)
		  )
	      do
		(feedback 1000)
		(incf row-count)
		)))
      
      (format t "~%[Processed ~A data rows from file ~A]~%" row-count filename)

      result
      )  
    )



(defun variable-column-row-to-list (row column-sep &key datatypes
							no-empty-columns-p 
							null-value
							first-column-sep
							trim-token-chars)
  (let (elements 
	;; The length of the raw row string in characters:
	row-length
	current-column
	current-type
	current-value
	(start 0) )
 
    (setq trim-token-chars (listify trim-token-chars))

    (handler-case
	(progn
#|	  (unless (if (consp column-sep)
		      (find (char row (1- (length row))) column-sep)
		    (eql (char row (1- (length row))) column-sep)
		    )
	    (setq row (concat row (string (if (consp column-sep)
					      (first column-sep)
					    column-sep))))
					    )
|#
	  (setq row-length (length row))
	  	  
	  ;; Returns T if char is a member of the current column separator set --
	  ;; that set could depend on whether we are on column 0 or later columns.
	  (flet ((column-separator-p (char &key first-column-p)
		   (let ((column-sep-set (if (and first-column-p
						  first-column-sep)
					     first-column-sep
					   column-sep)))
		     (if (consp column-sep-set)
			 (find char column-sep-set)
		       (eql char column-sep-set)))))

	  (loop
	      for i from 0
	      ;; taltman:Sep-2-2011 If we have alternate separators for first column,
	      ;; and we're processing the first column, use the alternate set only.
	      ;; Note end will be nil if sep is not found in remainder of line.
	      for end = (position-if #'(lambda (char) (column-separator-p char :first-column-p (= i 0))) row :start start)
			;; Use the n'th datatype. if we surpass the list of provided types,
			;; then reuse the last type:
	      for type = (when datatypes
			   (if (< i (length datatypes))
			       (nth i datatypes)
			     (nth (1- (length datatypes))
				  datatypes)))

			 ;; As we parse the line, this represents the raw string between two consecutive column separator characters:
	      for current-token = (subseq row start end)
				    
	      do
		(when trim-token-chars
		  ;; If trim chars for the tokens are specified, then trim
		  (setq current-token (string-trim trim-token-chars current-token)) )
		(setf current-value current-token)
		(setf current-column i)
		(setf current-type type)

		;; Coerce the value current-value into the proper type, if specified.  
		(handler-case
		    (if (string-equal "" current-token)
			(setf current-value null-value)
		      (when type
			(ecase type
			  (:string)
			  (:number (setq current-value (read-from-string current-value))
				   ;; This tries to do addition on the value of 'current-value'.
				   ;; If we didn't parse a number, this will fail:
				   (+ current-value 1))
			  (:symbol (setq current-value (intern current-value *package*)))
			  )))
		  (error ()		      
		    (error 'malformed-datafile-cell-error)))
		(push current-value elements)
		(when end
		  ;; Set start only when sep char was found in the line; if it was not found
		  ;; then there is no next token in this line.
		  (setq start (if no-empty-columns-p
				  ;;kr:Jan-23-2006 Skip all consecutive column-separators
				  (or (position-if-not #'(lambda (char) (column-separator-p char :first-column-p (= i 0))) row :start (1+ end))
				      ;; Skip at least the last separator.
				      (1+ end)
				      )
				;; Keep going after the separator.
				(1+ end)))
		  )
	      while (and end
			 (< start row-length))))
	  
	  (nreverse elements))
      (malformed-datafile-cell-error ()
	(warn "variable-column-row-to-list: unable to parse '~A' in column ~A~@[ as variable of type ~A~].~%"
	      current-value current-column current-type )
	nil))))

